package ru.dbotthepony.kstarbound.server.world

import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.arrays.Object2DArray
import ru.dbotthepony.kommons.io.BTreeDB
import ru.dbotthepony.kommons.io.ByteKey
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kommons.io.readVarInt
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.json.VersionedJson
import ru.dbotthepony.kstarbound.world.CHUNK_SIZE
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.api.AbstractCell
import ru.dbotthepony.kstarbound.world.api.ImmutableCell
import ru.dbotthepony.kstarbound.world.api.MutableCell
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import ru.dbotthepony.kstarbound.world.entities.WorldObject
import java.io.BufferedInputStream
import java.io.ByteArrayInputStream
import java.io.DataInputStream
import java.util.concurrent.CompletableFuture
import java.util.zip.Inflater
import java.util.zip.InflaterInputStream

class LegacyChunkSource(val db: BTreeDB<ByteKey, ByteArray>) : IChunkSource {
	override fun getTiles(pos: ChunkPos): CompletableFuture<KOptional<Object2DArray<out AbstractCell>>> {
		val chunkX = pos.x
		val chunkY = pos.y
		val key = ByteKey(1, (chunkX shr 8).toByte(), chunkX.toByte(), (chunkY shr 8).toByte(), chunkY.toByte())

		return db.read(key).thenApplyAsync {
			it.map {
				val reader = DataInputStream(BufferedInputStream(InflaterInputStream(ByteArrayInputStream(it), Inflater())))
				reader.skipBytes(3)

				val result = Object2DArray.nulls<ImmutableCell>(CHUNK_SIZE, CHUNK_SIZE)

				for (y in 0 until CHUNK_SIZE) {
					for (x in 0 until CHUNK_SIZE) {
						result[x, y] = MutableCell().read(reader).immutable()
					}
				}

				result as Object2DArray<out AbstractCell>
			}
		}
	}

	override fun getEntities(pos: ChunkPos): CompletableFuture<KOptional<Collection<AbstractEntity>>> {
		val chunkX = pos.x
		val chunkY = pos.y
		val key = ByteKey(2, (chunkX shr 8).toByte(), chunkX.toByte(), (chunkY shr 8).toByte(), chunkY.toByte())

		return db.read(key).thenApplyAsync {
			it.map {
				val reader = DataInputStream(BufferedInputStream(InflaterInputStream(ByteArrayInputStream(it), Inflater())))
				val i = reader.readVarInt()
				val objects = ArrayList<AbstractEntity>()

				for (i2 in 0 until i) {
					val obj = VersionedJson(reader)

					if (obj.identifier == "ObjectEntity") {
						try {
							objects.add(WorldObject.fromJson(obj.content.asJsonObject))
						} catch (err: Throwable) {
							LOGGER.error("Unable to deserialize entity in chunk $pos", err)
						}
					} else {
						LOGGER.error("Unknown entity type in chunk $pos: ${obj.identifier}")
					}
				}

				objects
			}
		}
	}

	companion object {
		private val vectors by lazy { Starbound.gson.getAdapter(Vector2i::class.java) }
		private val LOGGER = LogManager.getLogger()
	}
}
